package ru.dbotthepony.mc.otm.network

import io.netty.buffer.ByteBuf
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import net.minecraft.core.RegistryAccess
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.network.codec.StreamCodec
import ru.dbotthepony.kommons.io.readVarInt
import ru.dbotthepony.kommons.io.writeVarInt
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.NoSuchElementException

interface MatteryStreamCodec<S : ByteBuf, V> : StreamCodec<S, V> {
	fun copy(value: V): V {
		return value
	}

	fun compare(a: V, b: V): Boolean {
		return a == b
	}

	// ignore kotlin compiler warning; StreamCodec<> has wrong package-wide nullability annotation
	override fun decode(stream: S): V
	override fun encode(stream: S, value: V)

	class Wrapper<S : ByteBuf, V>(parent: StreamCodec<S, V>) : MatteryStreamCodec<S, V>, StreamCodec<S, V> by parent

	class Nullable<S : ByteBuf, V>(val parent: MatteryStreamCodec<S, V>) : MatteryStreamCodec<S, V?> {
		override fun decode(stream: S): V? {
			return if (!stream.readBoolean()) null else parent.decode(stream)
		}

		override fun encode(stream: S, value: V?) {
			if (value === null)
				stream.writeBoolean(false)
			else {
				stream.writeBoolean(true)
				parent.encode(stream, value)
			}
		}

		override fun copy(value: V?): V? {
			return if (value === null) null else parent.copy(value)
		}

		override fun compare(a: V?, b: V?): Boolean {
			if (a === null && b === null) return true
			if (a === null || b === null) return false
			return parent.compare(a, b)
		}
	}

	class Optional<S : ByteBuf, V : Any>(val parent: MatteryStreamCodec<S, V>) : MatteryStreamCodec<S, java.util.Optional<V>> {
		override fun decode(stream: S): java.util.Optional<V> {
			return if (!stream.readBoolean()) java.util.Optional.empty() else java.util.Optional.of(parent.decode(stream))
		}

		override fun encode(stream: S, value: java.util.Optional<V>) {
			if (value.isEmpty)
				stream.writeBoolean(false)
			else {
				stream.writeBoolean(true)
				parent.encode(stream, value.get())
			}
		}

		override fun copy(value: java.util.Optional<V>): java.util.Optional<V> {
			return value.map(parent::copy)
		}

		override fun compare(a: java.util.Optional<V>, b: java.util.Optional<V>): Boolean {
			if (a.isEmpty && b.isEmpty) return true
			if (a.isEmpty || b.isEmpty) return false
			return parent.compare(a.get(), b.get())
		}
	}

	class Collection<S : FriendlyByteBuf, E, C : MutableCollection<E>>(val elementCodec: MatteryStreamCodec<in S, E>, val collectionFactory: (Int) -> C) : MatteryStreamCodec<S, C> {
		override fun decode(stream: S): C {
			val size = stream.readVarInt()

			if (size <= 0) {
				return collectionFactory.invoke(0)
			}

			val collection = collectionFactory.invoke(size)

			for (i in 0 until size) {
				collection.add(elementCodec.decode(stream))
			}

			return collection
		}

		override fun encode(stream: S, value: C) {
			stream.writeVarInt(value.size)
			value.forEach { elementCodec.encode(stream, it) }
		}

		override fun copy(value: C): C {
			val new = collectionFactory.invoke(value.size)
			value.forEach { new.add(elementCodec.copy(it)) }
			return new
		}
	}

	class Enum<S : FriendlyByteBuf, V : kotlin.Enum<V>>(clazz: Class<out V>) : MatteryStreamCodec<S, V> {
		val clazz = searchClass(clazz)
		val values: List<V> = listOf(*this.clazz.enumConstants!!)
		val valuesMap = values.associateBy { it.name }

		override fun decode(stream: S): V {
			val id = stream.readVarInt()
			return values.getOrNull(id) ?: throw NoSuchElementException("No such enum with index $id")
		}

		override fun encode(stream: S, value: V) {
			stream.writeVarInt(value.ordinal)
		}

		override fun copy(value: V): V {
			return value
		}

		override fun compare(a: V, b: V): Boolean {
			return a === b
		}

		companion object {
			/**
			 * FIXME: enums with abstract methods which get compiled to subclasses, whose DO NOT expose "parent's" enum constants array
			 *
			 * is there an already existing solution?
			 */
			fun <V : kotlin.Enum<V>> searchClass(clazz: Class<out V>): Class<out V> {
				var search: Class<*> = clazz

				while (search.enumConstants == null && search.superclass != null) {
					search = search.superclass
				}

				if (search.enumConstants == null) {
					throw ClassCastException("$clazz does not represent an enum or enum subclass")
				}

				return search as Class<out V>
			}
		}
	}
}
